package eu.jonahbauer.json;

import eu.jonahbauer.json.token.JsonTokenKind;
import eu.jonahbauer.json.token.impl.ComplexTokenKind;
import eu.jonahbauer.json.token.JsonToken;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * Java representation of a JSON boolean. Note, that JSON does not distinguish between integers and floating point
 * numbers and therefore all numbers are stored as {@code double}.
 */
public record JsonNumber(double value) implements JsonValue, JsonToken {
    public static final JsonTokenKind KIND = ComplexTokenKind.NUMBER;
    private static final long LONG_BIT_MASK = (~0L << Double.PRECISION);

    public JsonNumber {
        if (!Double.isFinite(value)) throw new IllegalArgumentException("value must be finite");
    }

    /**
     * Creates a {@link JsonNumber} wrapping the given {@code int} value.
     * @param i an {@code int} value
     * @return a {@link JsonNumber}
     */
    public static @NotNull JsonNumber valueOf(int i) {
        return new JsonNumber(i);
    }

    /**
     * Creates a {@link JsonNumber} wrapping the given {@code long} value.
     * @param l a {@code long} value
     * @return a {@link JsonNumber}
     * @throws IllegalArgumentException if conversion from {@code long} to {@code double} is lossy
     */
    public static @NotNull JsonNumber valueOf(long l) {
        return new JsonNumber(checkLongRange(l));
    }

    /**
     * Creates a {@link JsonNumber} wrapping the given {@code double} value.
     * @param d an {@code double} value
     * @return a {@link JsonNumber}
     * @throws IllegalArgumentException if {@code d} is not {@linkplain Double#isFinite(double) finite}
     */
    public static @NotNull JsonNumber valueOf(double d) {
        return new JsonNumber(d);
    }

    /**
     * Creates a {@link JsonNumber} wrapping {@code number}. {@code number} is converted to a {@code double} using
     * {@link Number#doubleValue()}.
     * @param number a number
     * @return a {@link JsonNumber}
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonNumber valueOf(@Nullable Number number) {
        return number == null ? null : new JsonNumber(number.doubleValue());
    }

    private static long checkLongRange(long value) {
        if (((value >= 0 ? value : -value) & LONG_BIT_MASK) != 0) {
            throw new IllegalArgumentException("lossy conversion from long to double for value " + value);
        }
        return value;
    }

    @Override
    public @NotNull String toString() {
        return Double.toString(value);
    }

    @Override
    public @NotNull JsonTokenKind getKind() {
        return KIND;
    }
}
