package eu.jonahbauer.json.parser.tokenizer;

import eu.jonahbauer.json.token.*;
import lombok.Getter;
import eu.jonahbauer.json.JsonBoolean;
import eu.jonahbauer.json.JsonNumber;
import eu.jonahbauer.json.JsonString;
import eu.jonahbauer.json.parser.JsonTokenizerException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.*;
import java.util.*;
import java.util.regex.Pattern;

public class JsonTokenizerImpl implements JsonTokenizer {
    private static final Pattern NUMBER_PATTERN = Pattern.compile("-?(?:0|[1-9]\\d*)(?:\\.\\d+)?([eE][+-]?\\d+)?");
    private final @NotNull TemplateReader reader;
    private final int[] buffer = new int[4];

    @Getter
    private int fragment;
    @Getter
    private int line;
    @Getter
    private int column;

    public JsonTokenizerImpl(@NotNull String json) {
        this.reader = new TemplateReader(json);
    }

    public JsonTokenizerImpl(@NotNull Reader reader) {
        this.reader = new TemplateReader(reader);
    }

    public JsonTokenizerImpl(@NotNull StringTemplate json) {
        this.reader = new TemplateReader(json);
    }

    @Override
    public @Nullable JsonToken next() throws IOException {
        int chr;

        do {
            chr = reader.read();
        } while (isWhitespace(chr));

        fragment = reader.getFragment();
        line = reader.getLine();
        column = reader.getColumn();

        return switch (chr) {
            case TemplateReader.EOF -> null;
            case TemplateReader.PLACEHOLDER -> new JsonPlaceholder(reader.getObject());
            case '{' -> JsonPunctuation.BEGIN_OBJECT;
            case '}' -> JsonPunctuation.END_OBJECT;
            case '[' -> JsonPunctuation.BEGIN_ARRAY;
            case ']' -> JsonPunctuation.END_ARRAY;
            case ',' -> JsonPunctuation.VALUE_SEPARATOR;
            case ':' -> JsonPunctuation.NAME_SEPARATOR;
            case '"' -> {
                StringTemplate string = readString();
                if (string.fragments().size() == 1) {
                    yield new JsonString(string.fragments().getFirst());
                } else {
                    yield new JsonStringTemplate(string);
                }
            }
            default -> {
                if (chr < 0) throw new AssertionError();

                reader.pushback();
                String token = nextToken();
                yield switch (token) {
                    case "true" -> JsonBoolean.TRUE;
                    case "false" -> JsonBoolean.FALSE;
                    case "null" -> JsonNull.NULL;
                    case String number when isNumberLiteral(number) -> {
                        try {
                            yield new JsonNumber(Double.parseDouble(number));
                        } catch (IllegalArgumentException ex) {
                            throw error("invalid number literal: " + number, ex);
                        }
                    }
                    default -> throw error("invalid token: " + token);
                };
            }
        };
    }

    @Override
    public boolean hasTemplateSupport() {
        return reader.isHasTemplateSupport();
    }

    protected boolean isWhitespace(int chr) {
        return chr == ' ' || chr == '\t' || chr == '\r' || chr == '\n';
    }

    protected boolean isPunctuation(int chr) {
        return chr == '{' || chr == '}' || chr == '[' || chr == ']' || chr == ',' || chr == ':' || chr == '"';
    }

    protected boolean isNumberLiteral(@NotNull String token) {
        return NUMBER_PATTERN.matcher(token).matches();
    }

    private @NotNull String nextToken() throws IOException {
        StringBuilder out = new StringBuilder();
        while (true) {
            int chr = reader.read();
            if (isWhitespace(chr) || isPunctuation(chr)) {
                reader.pushback();
                return out.toString();
            } else if (chr == -1) {
                return out.toString();
            } else {
                out.append((char) chr);
            }
        }
    }

    private @NotNull StringTemplate readString() throws IOException {
        List<String> fragments = new ArrayList<>();
        List<Object> values = new ArrayList<>();

        StringBuilder current = new StringBuilder();
        while (true) {
            int chr = reader.read();
            if (chr == '"') {
                fragments.add(current.toString());
                return StringTemplate.of(fragments, values);
            } else if (chr == '\\') {
                chr = reader.read();
                switch (chr) {
                    case TemplateReader.EOF -> throw error("incomplete escape sequence in string literal");
                    case '"' -> current.append('"');
                    case '\\' -> current.append('\\');
                    case '/' -> current.append('/');
                    case 'b' -> current.append('\b');
                    case 'f' -> current.append('\f');
                    case 'n' -> current.append('\n');
                    case 'r' -> current.append('\r');
                    case 't' -> current.append('\t');
                    case 'u' -> {
                        buffer[0] = reader.read();
                        buffer[1] = reader.read();
                        buffer[2] = reader.read();
                        buffer[3] = reader.read();

                        if (buffer[0] < 0 || buffer[1] < 0 || buffer[2] < 0 || buffer[3] < 0) {
                            throw error("incomplete escape sequence in string literal");
                        }

                        int code = 0;
                        for (int i = 0; i < 4; i++) {
                            code *= 16;
                            char esc = (char) buffer[i];
                            if ('0' <= esc && esc <= '9') {
                                code += esc - '0';
                            } else if ('a' <= esc && esc <= 'f') {
                                code += 11 + (esc - 'a');
                            } else if ('A' <= esc && esc <= 'F') {
                                code += 11 + (esc - 'A');
                            } else {
                                throw error(STR."invalid character \{toString(buffer[i])} in escape sequence");
                            }
                        }
                        current.append((char) code);
                    }
                    default -> throw error(STR."invalid character \{toString(chr)} in escape sequence");
                }

            } else if (chr == TemplateReader.EOF) {
                throw error("unclosed string literal");
            } else if (chr == TemplateReader.PLACEHOLDER) {
                fragments.add(current.toString());
                current.setLength(0);
                values.add(reader.getObject());
            } else if (chr < 32) {
                throw error("unescaped control character in string literal");
            } else {
                current.append((char) chr);
            }
        }
    }

    private @NotNull JsonTokenizerException error(@NotNull String message) {
        return new JsonTokenizerException(getFragment(), getLine(), getColumn(), message);
    }

    private @NotNull JsonTokenizerException error(@NotNull String message, @NotNull Throwable cause) {
        return new JsonTokenizerException(getFragment(), getLine(), getColumn(), message, cause);
    }

    private static @NotNull String toString(int chr) {
        return switch (chr) {
            case TemplateReader.EOF -> "EOF";
            case TemplateReader.PLACEHOLDER -> "PLACEHOLDER";
            case '\t' -> "'\\t'";
            case '\r' -> "'\\r'";
            case '\n' -> "'\\n'";
            case '\b' -> "'\\b'";
            case '\f' -> "'\\f'";
            default -> {
                if (Character.isISOControl((char) chr)) {
                    yield STR."0x\{Integer.toHexString(chr)}";
                } else {
                    yield STR."'\{(char) chr}'";
                }
            }
        };
    }

}
