package eu.jonahbauer.json.parser.tokenizer;

import lombok.Getter;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.VisibleForTesting;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.List;
import java.util.Objects;
import java.util.function.IntFunction;

@VisibleForTesting
class TemplateReader {
    public static final int EOF = -1;
    public static final int PLACEHOLDER = -2;


    private final IntFunction<Reader> readers;
    private final List<Object> values;
    private final int length;

    @Getter
    private final boolean hasTemplateSupport;

    private Reader reader;

    private int fragment;
    @Getter
    private int line = 1;
    @Getter
    private int column;


    private int current;
    private Object currentObject;

    private boolean pushback;


    public TemplateReader(@NotNull String string) {
        this(new StringReader(string));
    }

    public TemplateReader(@NotNull Reader reader) {
        Objects.requireNonNull(reader, "reader");
        this.readers = _ -> reader;
        this.values = List.of();
        this.length = 1;
        this.hasTemplateSupport = false;
    }

    public TemplateReader(@NotNull StringTemplate template) {
        Objects.requireNonNull(template, "template");
        this.readers = i -> new StringReader(template.fragments().get(i));
        this.values = template.values();
        this.length = template.fragments().size();
        this.hasTemplateSupport = true;
    }

    public int read() throws IOException {
        if (pushback) {
            pushback = false;
            return current;
        }

        if (!ensureReader()) return EOF;

        int result = reader.read();

        // track position
        if (current != '\r' && result == '\n' || result == '\r') {
            line++;
            column = 0;
        } else if (result != '\n') {
            column++;
        }

        if (result == EOF) {
            reader = null;
            if (fragment < length) {
                currentObject = values.get(fragment - 1);
                current = PLACEHOLDER;
            } else {
                currentObject = null;
                current = result;
            }
        } else {
            currentObject = null;
            current = result;
        }

        return current;
    }

    private boolean ensureReader() {
        if (reader != null) {
            return true;
        } else if (fragment < length) {
            reader = readers.apply(fragment++);
            line = 1;
            column = 0;
            return true;
        } else {
            return false;
        }
    }

    public Object getObject() {
        if (current != PLACEHOLDER) {
            throw new IllegalStateException();
        }
        return currentObject;
    }

    public void pushback() {
        if (pushback) throw new IllegalStateException();
        pushback = true;
    }

    public int getFragment() {
        return hasTemplateSupport ? fragment : -1;
    }
}
